import numpy as np

from genetist.individual import Individual

class Mutation:
    def __init__(self, mutation_type: str, prob_mutation: float, search_space_type: str, params: dict):
        self.mutation_type = mutation_type
        self.prob_mutation = prob_mutation
        self.search_space_type = search_space_type
        self.params = params
    
    def _fixed_search_mutation_process(self, child: Individual, gene_index: int, param: str) -> Individual:
        child_genome = child.get_genome()
        if len(self.params.get(param)) == 2:
            if self.params.get(param)[0] != child_genome[gene_index]:
                child_genome[gene_index] = self.params.get(param)[0]
            else:
                child_genome[gene_index] = self.params.get(param)[1]
        else:
            child_genome[gene_index] = np.random.choice(self.params.get(param))
        
        child.set_genome(child_genome)
        
        return child

    def _flexible_search_mutation_process(self, child: Individual, gene_index: int, param: str) -> Individual:
        child_genome = child.get_genome()
        if self.params.get(param).get('type') == 'int':
            if self.params[param]['low'] == 0 and self.params[param]['high'] == 1:
                 if child_genome[gene_index] == 0:
                    child_genome[gene_index] = 1
                 else:
                    child_genome[gene_index] = 0
            else:
                child_genome[gene_index] = np.random.random_integers(self.params[param]['low'], (self.params[param]['high']))
        elif self.params.get(param).get('type') == 'float':
            child_genome[gene_index] = np.random.uniform(self.params[param]['low'], self.params[param]['high'])
        elif self.params.get(param).get('type') == 'categorical':
            if len(self.params.get(param).get('choices')) == 2:
                if self.params.get(param).get('choices')[0] != child_genome[gene_index]:
                    child_genome[gene_index] = self.params.get(param).get('choices')[0]
                else:
                    child_genome[gene_index] = self.params.get(param).get('choices')[1]
            else:
                child_genome[gene_index] = np.random.choice(self.params.get(param).get('choices'))

        child.set_genome(child_genome)
        
        return child

    def _single_mutation_in_fixed_search(self, child: Individual) -> Individual:
        gene_index = np.random.randint(0,  len(child))
        param = list(self.params.keys())[gene_index]
        child = self._fixed_search_mutation_process(child, gene_index, param)
            
        return child

    def _single_mutation_in_flexible_search(self, child: Individual) -> Individual:
        gene_index = np.random.randint(0,  len(child))
        param = list(self.params.keys())[gene_index]
        child = self._flexible_search_mutation_process(child, gene_index, param)
        
        return child

    def _multiple_mutation_in_fixed_search(self, child: Individual) -> Individual:
        number_of_mutations = np.random.randint(1, len(child))
        gene_indexes = np.random.randint(0, len(child), size=number_of_mutations)
        for gene_index in gene_indexes:
            param = list(self.params.keys())[gene_index]
            child = self._fixed_search_mutation_process(child, gene_index, param)
        
        return child
                

    def _multiple_mutation_in_flexible_search(self, child: Individual) -> Individual:
        number_of_mutations = np.random.randint(1, len(child))
        gene_indexes = np.random.randint(0, len(child), size=number_of_mutations)
        for gene_index in gene_indexes:
            param = list(self.params.keys())[gene_index]
            child = self._flexible_search_mutation_process(child, gene_index, param)
        
        return child

    def mutate(self, child: Individual) -> Individual:
        if np.random.rand() < self.prob_mutation:
            if self.mutation_type == 'single_gene':
                if self.search_space_type == 'fixed_search':
                    return self._single_mutation_in_fixed_search(child)
                else:
                    return self._single_mutation_in_flexible_search(child)
            elif self.mutation_type == 'multiple_genes':
                if self.search_space_type == 'fixed_search':
                    return self._multiple_mutation_in_fixed_search(child)
                else:
                    return self._multiple_mutation_in_flexible_search(child)
            else:
                raise Exception(f'Mutation type {self.mutation_type} not supported.')
        else:
            return child


import pandas as pd
import datetime

from typing import Union

class Results:
    def __init__(self):
        self.best_score = None
        self.best_individual = None
        self.execution_time = None
        self.best_per_generation_dataframe = pd.DataFrame()
    
    def get_best_score(self) -> None:
        self.best_score = self.best_per_generation_dataframe['best_score'].max()

    def get_best_individual(self) -> None:
        best_generation = self.best_per_generation_dataframe[self.best_per_generation_dataframe['best_score'] == self.best_per_generation_dataframe['best_score'].max()]
        self.best_individual = best_generation.iloc[0,2:].to_dict()
    
    def set_execution_time(self, execution_time: float) -> None:
        self.execution_time = execution_time
        time = str(datetime.timedelta(seconds=self.execution_time))
        hours = time.split(':')[0]
        minutes = time.split(':')[1]
        seconds = time.split(':')[2]
        self.execution_time = f'{hours} hours {minutes} minutes {seconds} seconds'
    
    def add_generation_results(self, generation: int, best_score: Union[int, float], best_individual: dict) -> None:
            generation_results = {'generation': generation, 'best_score': best_score}
            generation_results.update(best_individual)
            values = [list(generation_results.values())]
            columns = list(generation_results.keys())
            df_generation_results = pd.DataFrame(values, columns=columns)
            self.best_per_generation_dataframe = pd.concat([self.best_per_generation_dataframe, df_generation_results], axis=0)

    def sort_best_per_generation_dataframe(self, column: str, direction: str) -> None:
        if direction == 'maximize':
            self.best_per_generation_dataframe = self.best_per_generation_dataframe.sort_values(by=column, ascending=False)
        elif direction == 'minimize':
            self.best_per_generation_dataframe = self.best_per_generation_dataframe.sort_values(by=column, ascending=True)
        else:
            raise Exception(f'Direction {direction} not supported.')


import numpy as np
import random
import time
import math
import logging

from typing import Callable, List, Tuple, Union
from multiprocessing import Pool, cpu_count
from tqdm import tqdm
from genetist.crossover import Crossover
from genetist.mutation import Mutation
from genetist.individual import Individual
from genetist.datatype_inference import DataTypeInference
from genetist.results import Results

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger('ENVIRONMENT')

class Environment:
    def __init__(self, params: dict, num_population: int = 100, generations: int = 100, crossover_type: str = 'one_point', mutation_type: str = 'single_gene', prob_mutation: float = 0.1, elite_rate: float = 0.1, verbose: int = 1):
        self.params = params
        self.num_population = num_population
        self.generations = generations
        self.crossover_type = crossover_type
        self.mutation_type = mutation_type
        self.prob_mutation = prob_mutation
        self.elite_rate = elite_rate
        self.verbose = verbose

        self.search_space_type = DataTypeInference.infer_search_space_type(params)
        self.crossover = Crossover(self.crossover_type)
        self.mutation = Mutation(self.mutation_type, self.prob_mutation, self.search_space_type, self.params)
            
    def _initialize_population(self, objective: Callable[[dict], Union[int,float]]) -> List[Individual]:
        if self.verbose > 1: logger.info(f'Initializing population...')
        population = list()
        for _ in range(self.num_population):
            population.append(Individual(self.params, self.search_space_type, objective))

        return population
    
    def _calculate_fitness_process(self, individual: Individual) -> Individual:
        individual.calculate_fitness()
        return individual

    def _calculate_population_fitness(self, individuals: List[Individual], n_jobs: int = 1) -> List[Individual]:
        if self.verbose > 1: logger.info(f'Calculating population fitness...')
        if n_jobs == -1: n_jobs = cpu_count()

        if n_jobs != 1:
            pool = Pool(n_jobs)
            new_individuals = pool.map(self._calculate_fitness_process, individuals)
        else:
            new_individuals = list()
            for individual in individuals:
                new_individuals.append(self._calculate_fitness_process(individual))

        return new_individuals

    def _order_population_by_fitness(self, individuals: List[Individual], direction: str) -> List[Individual]:
        if self.verbose > 1: logger.info(f'Ordering population by fitness...')
        if direction == 'maximize':
            individuals = sorted(individuals, key=lambda individual: individual.get_fitness(), reverse=True)
        elif direction == 'minimize':
            individuals = sorted(individuals, key=lambda individual: individual.get_fitness())
        else:
            raise Exception(f'Direction {direction} not supported.')

        return individuals

    def _choose_parents(self, individuals: List[Individual]) -> List[Tuple[Individual, Individual]]:
        if self.verbose > 1: logger.info(f'Selecting parents...')
        parents = list()
        weights = np.arange(len(individuals), 0, step=-1)
        number_of_parents = math.ceil(len(individuals) * (1 - self.elite_rate)) // 2
        for _ in range(number_of_parents):
            parents.append(
                random.choices(
                    population=individuals,
                    weights=weights,
                    k=2
                )
            )

        return parents

    def _run_crossover_with_mutation(self, best_parents: List[Tuple[Individual, Individual]]) -> List[Individual]:
        if self.verbose > 1: logger.info(f'Running crossover with mutation...')
        childs = list()
        for parents in best_parents:
            child_1, child_2 = self.crossover.crossover(parents[0], parents[1])
            child_1 = self.mutation.mutate(child_1)
            child_2 = self.mutation.mutate(child_2)
            childs.append(child_1)
            childs.append(child_2)

        return childs

    def _get_elite(self, individuals: List[Individual]) -> List[Individual]:
        if self.verbose > 1: logger.info(f'Getting elite individuals...')
        elite = individuals[:math.ceil(len(individuals) * self.elite_rate)]

        return elite
    
    def optimize(self, objective: Callable[[dict], Union[int,float]], direction: str, n_jobs: int = 1) -> Results:
        start_time = time.time()
        
        results = Results()
        individuals = self._initialize_population(objective)
        individuals = self._calculate_population_fitness(individuals, n_jobs)
        individuals = self._order_population_by_fitness(individuals, direction)
        progress_bar = tqdm(range(0, self.generations))
        for generation in progress_bar:
            elite_individuals = self._get_elite(individuals)
            parents = self._choose_parents(individuals)
            individuals = self._run_crossover_with_mutation(parents)
            individuals.extend(elite_individuals)
            individuals = self._calculate_population_fitness(individuals, n_jobs)
            individuals = self._order_population_by_fitness(individuals, direction)
            best_individual = individuals[0].get_name_genome_genes()
            best_score = individuals[0].get_fitness()

            results.add_generation_results(generation+1, best_score, best_individual)
            progress_bar.set_description(f'RUNNING GENERATION {generation + 1}')
            if self.verbose >= 1:
                logger.info(f'THE BEST SOLUTION IN GENERATION {generation+1} IS {best_individual} WITH A SCORE OF {best_score}')

        end_time = time.time()

        results.set_execution_time(end_time - start_time)
        results.sort_best_per_generation_dataframe(column='best_score', direction=direction)
        results.get_best_score()
        results.get_best_individual()
            
        return results




import numpy as np

from typing import Callable, Union

class Individual:
    def __init__(self, params: dict, search_space_type: str, objective: Callable[[dict], Union[int,float]]):
        self.params = params
        self.search_space_type = search_space_type
        self.objective = objective
        self.genome = self._initialize_genome()
        self.fitness = None
    
    def __len__(self) -> int:
        return len(self.genome)
    
    def _initialize_genome(self) -> list:
        genome = list()
        if self.search_space_type == 'flexible_search':
            for _, values in self.params.items():
                if values['type'] == 'int':
                    genome.append(np.random.random_integers(values['low'], values['high']))
                elif values['type'] == 'float':
                    genome.append(np.random.uniform(values['low'], values['high']))
                elif values['type'] == 'categorical':
                    genome.append(np.random.choice(values['choices']))
                else:
                    raise ValueError(f'Type {values["type"]} not supported.')
        else:
            for _, values in self.params.items():
                genome.append(np.random.choice(values))
        
        return genome
    
    def get_name_genome_genes(self) -> dict:
        genome_gene_names = {}
        names = list(self.params.keys())
        for name, gene in zip(names, self.genome):
            genome_gene_names[name] = gene
        
        return genome_gene_names
    
    def calculate_fitness(self) -> None:
        self.fitness = self.objective(self.get_name_genome_genes())
    
    def get_genome(self) -> list:
        return self.genome
    
    def get_fitness(self) -> Union[int, float]:
        return self.fitness
    
    def set_genome(self, genome: list) -> None:
        self.genome = genome


from typing import List

class Params:

    @staticmethod
    def suggest_categorical(choices: List[str]) -> dict:
        for choice in choices:
            if isinstance(choice, str) == False:
                raise Exception(f'Choices must be string.')
        return {'type': 'categorical', 'choices': choices}

    @staticmethod
    def suggest_int(low: int, high: int) -> dict:
        if isinstance(low, int) == False and isinstance(low, float) == False:
            raise Exception(f'Low must be a number.')
        elif isinstance(high, int) == False and isinstance(high, float) == False:
            raise Exception(f'High must be a number.')
        else:
            return {'type': 'int', 'low': low, 'high': high}

    @staticmethod
    def suggest_float(low: float, high: float) -> dict:
        if isinstance(low, float) == False and isinstance(low, int) == False:
            raise Exception(f'Low must be float.')
        elif isinstance(high, float) == False and isinstance(high, int) == False:
            raise Exception(f'High must be a number.')
        else:
            return {'type': 'float', 'low': low, 'high': high}


import numpy as np

class DataTypeInference:

    @staticmethod
    def infer_search_space_type(params: dict) -> str:
        key = list(params.keys())[0]
        if isinstance(params[key], dict):
            search_space_type = 'flexible_search'
        elif isinstance(params[key], list) or isinstance(params[key], np.ndarray):
            search_space_type = 'fixed_search'
        else:
            raise Exception('Unable to infer search space type. Please check your search space format.')
    
        return search_space_type


from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

INSTALL_REQUIRES = [
    'numpy==1.21.3',
    'tqdm==4.64.0',
    'pandas==1.3.4',
]

VERSION = '0.9.9'

if __name__ == '__main__':
    setup(
        name='genetist',
        description='Genetist: optimization with genetic algorithms',
        long_description=long_description,
        version=VERSION,
        long_description_content_type='text/markdown',
        author='Unai Lopez Ansoleaga',
        author_email='unai19970315@gmail.com',
        url='https://github.com/unaiLopez/genetist',
        license='MIT',
        packages=['genetist'],
        include_package_data=True,
        install_requires=INSTALL_REQUIRES,
        python_requires='>=3.6'
    )



import numpy as np

from copy import copy
from typing import Tuple
from genetist.individual import Individual

class Crossover:
    def __init__(self, crossover_type: str):
        self.crossover_type = crossover_type

    def _one_point_crossover(self, parent_1: Individual, parent_2: Individual) -> Tuple[Individual, Individual]:
        point = np.random.randint(1, len(parent_1) - 1)
        
        child_1_genome = parent_1.get_genome()[:point] + parent_2.get_genome()[point:]
        child_2_genome = parent_2.get_genome()[:point] + parent_1.get_genome()[point:]
        child_1 = copy(parent_1)
        child_2 = copy(parent_2)
        child_1.set_genome(child_1_genome)
        child_2.set_genome(child_2_genome)
        
        return child_1, child_2
        
    def _two_point_crossover(self, parent_1: Individual, parent_2: Individual) -> Tuple[Individual, Individual]:
        max_points = len(parent_1) - 2
        if max_points < 2:
            raise Exception(f'Too much points tried. Try less points for the crossover.')
        else:     
            points = sorted(np.random.randint(1, len(parent_1) - 1, size=2))
            point_1 = points[0]
            point_2 = points[1]
            
            child_1_genome = parent_1.get_genome()[:point_1] + parent_2.get_genome()[point_1:point_2] + parent_1.get_genome()[point_2:]
            child_2_genome = parent_2.get_genome()[:point_1] + parent_1.get_genome()[point_1:point_2] + parent_2.get_genome()[point_2:]

            child_1 = copy(parent_1)
            child_2 = copy(parent_2)
            child_1.set_genome(child_1_genome)
            child_2.set_genome(child_2_genome)
        
        return child_1, child_2
    
    def _three_point_crossover(self, parent_1: Individual, parent_2: Individual) -> Tuple[Individual, Individual]:
        max_points = len(parent_1) - 2
        if max_points < 3:
            raise Exception(f'Too much points tried. Try less points for the crossover.')
        else:
            points = sorted(np.random.randint(1, len(parent_1) - 1, size=3))
            point_1 = points[0]
            point_2 = points[1]
            point_3 = points[2]

            child_1_genome = parent_1.get_genome()[:point_1] + parent_2.get_genome()[point_1:point_2] + parent_1.get_genome()[point_2:point_3] +  parent_2.get_genome()[point_3:]
            child_2_genome = parent_2.get_genome()[:point_1] + parent_1.get_genome()[point_1:point_2] + parent_2.get_genome()[point_2:point_3] +  parent_1.get_genome()[point_3:]

            child_1 = copy(parent_1)
            child_2 = copy(parent_2)
            child_1.set_genome(child_1_genome)
            child_2.set_genome(child_2_genome)
        
        return child_1, child_2
        

    def crossover(self, parent_1: Individual, parent_2: Individual) -> Tuple[Individual, Individual]:
        if self.crossover_type == 'one_point':
            return self._one_point_crossover(parent_1, parent_2)
        elif self.crossover_type == 'two_point':
            return self._two_point_crossover(parent_1, parent_2)
        elif self.crossover_type == 'three_point':
            return self._three_point_crossover(parent_1, parent_2)
        else:
            raise Exception(f'Crossover {self.crossover_type} not supported.')
